'''
Create a datasource for a ClicRDV instance.

It is a gateway to http://developers.clicrdv.com/fr/rest-api.html

How to use in Passerelle:

* list of all sets :
  input : https//passerelle/data/clicrdv-title/?method=get_interventionsets
  output :
  { data: [ { id: 8762, text: "appointment for a passeport" },
          { id: 7044, text: "appointment for a id card" } ], err: 0 }

* list of interventions in a set
  input : https//passerelle/data/clicrdv-title/?method=get_interventions&set=7044
  output :
  { data: [ { id: 887, text: "for one person" },
          { id: 889, text: "for two persons" } ], err: 0 }

* list of free dates in a intervention
  input : https//passerelle/data/clicrdv-title/?method=get_dates&intervention=887
  output :
  { data: [ { id: '2014-05-07', text: "7 mai 2014" },
          { id: '2014-05-13', text: "13 mai 2014" } ], err: 0 }

* list of free time in a date, for an intervention
  input : https//passerelle/data/clicrdv-title/?method=get_times&date=2014-05-07&intervention=887
  output :
  { data: [ { id: '15:10:00', text: "15:10" },
          { id: '15:30:00', text: "15:30" } ], err: 0 }

* list of free date&times in a intervention
  input : https//passerelle/data/clicrdv-title/?method=get_datetimes&intervention=887

* create an appointment
  input (POST : URL and payload)
    https//passerelle/data/clicrdv-title/?method=create[&intervention=887][&websource=Web]
    { 'fields': {
        'clicrdv_intervention_raw': ..., # optional if present in URL
        'clicrdv_firstname': ...,
        'clicrdv_lastname': ...,
        'clicrdv_email': ...,
        'clicrdv_firstphone': ...,
        'clicrdv_secondphone': ...,
        'clicrdv_date_raw': ..., 'clicrdv_time_raw': ..., # or 'clicrdv_datetime_raw': ...,
        'clicrdv_comments': ..., # optional
        'clicrdc_fiche_xxx': ... # add a optional "xxx" value in fiche
    }
  output:
    { data: { 'success': true, 'appointment_id': 123 }, err: 0 }

* cancel an appointment :
  input : https//passerelle/data/clicrdv-title/?method=cancel&id=...
  output : { data: { 'success': true }, err: 0 }
  
'''

import base64
import datetime
import json
import urllib2

from django.conf import settings
from django.db import models
from django.utils.dateformat import format as date_format
from django.utils.dateformat import time_format

from passerelle.datasources.models import BaseDataSource

CLICRDV_SERVERS = (
    ('www.clicrdv.com', 'Production (www.clicrdv.com)'),
    ('sandbox.clicrdv.com', 'SandBox (sandbox.clicrdv.com)')
)


class ClicRdv(BaseDataSource):
    server = models.CharField(max_length=64,
        choices=CLICRDV_SERVERS,
        default='sandbox.clicrdv.com')
    apikey = models.CharField(max_length=64)
    username = models.CharField(max_length=64)
    password = models.CharField(max_length=64)
    websource = models.CharField(max_length=64, blank=True, null=True)
    default_comment = models.CharField(max_length=250, blank=True, null=True)

    parameters = ('method', 'set', 'intervention', 'date', 'websource', 'id')


    def get_request(self, uri):
        url = 'https://%s/api/v1/%s' % (self.server, uri)
        if '?' in url:
           url = url + '&apikey=%s&format=json' % self.apikey
        else:
           url = url + '?apikey=%s&format=json' % self.apikey
        req = urllib2.Request(url)
        authheader = 'Basic ' + \
            base64.encodestring('%s:%s' % (self.username, self.password))[:-1]
        req.add_header('Authorization', authheader)
        return req

    def get_json(self, uri):
        req = self.get_request(uri)
        return json.load(urllib2.urlopen(req))

    def get_interventionsets(self, **kwargs):
        records = self.get_json('interventionsets').get('records')
        records.sort(lambda x,y: cmp(x['sort'], y['sort']))
        ret = []
        for record in records:
            if record.get('publicname'):
                ret.append({'id': record['id'], 'text': record['publicname'], 'details': record})
        return ret

    def get_interventions(self, set, **kwargs):
        interventions_set = [int(set)]
        ret = []
        for iset in interventions_set:
            records = self.get_json('interventionsets/%s/interventions' % iset).get('records')
            records.sort(lambda x,y: cmp(x['sort'], y['sort']))
            for record in records:
                if record.get('publicname'):
                    ret.append({'id': record['id'], 'text': record['publicname'], 'details': record})
        return ret

    def get_available_timeslots(self, intervention, date_start=None, date_end=None):
        timeslots = []
        iid = int(intervention)
        gid = self.get_json('interventions/%d' % iid).get('group_id')
        request_uri = 'availabletimeslots?intervention_ids[]=%s&group_id=%s' % (iid, gid)
        if date_start is None:
            date_start = datetime.datetime.today().strftime('%Y-%m-%d')
        if date_end is None:
            date_end = (datetime.datetime.today() + datetime.timedelta(366)).strftime('%Y-%m-%d')
        if date_start:
            request_uri = request_uri + '&start=%s' % urllib2.quote(date_start)
        if date_end:
            request_uri = request_uri + '&end=%s' % urllib2.quote(date_end)
        for timeslot in self.get_json(request_uri).get('availabletimeslots'):
            timeslots.append(timeslot.get('start'))
        timeslots.sort()
        return timeslots

    def get_datetimes(self, intervention, **kwargs):
        datetimes = []
        for timeslot in self.get_available_timeslots(intervention):
            parsed = datetime.datetime.strptime(timeslot, '%Y-%m-%d %H:%M:%S')
            datetimed = {'id': parsed.strftime('%Y-%m-%d-%H:%M:%S'),
                        'text': date_format(parsed, 'j F Y H:i')}
            datetimes.append(datetimed)
        datetimes.sort(lambda x,y: cmp(x.get('id'), y.get('id')))
        return datetimes

    def get_dates(self, intervention, **kwargs):
        dates = []
        for timeslot in self.get_available_timeslots(intervention):
            parsed = datetime.datetime.strptime(timeslot, '%Y-%m-%d %H:%M:%S')
            date = {'id': parsed.strftime('%Y-%m-%d'),
                    'text': date_format(parsed, 'j F Y')}
            if date in dates:
                continue
            dates.append(date)
        dates.sort(lambda x,y: cmp(x.get('id'), y.get('id')))
        return dates

    def get_times(self, intervention, date, **kwargs):
        if not date:
            raise Exception('no date value')
        times = []
        for timeslot in self.get_available_timeslots(intervention,
                                date_start='%s 00:00:00' % date,
                                date_end='%s 23:59:59' % date):
            parsed = datetime.datetime.strptime(timeslot, '%Y-%m-%d %H:%M:%S')
            timed = {'id': parsed.strftime('%H:%M:%S'),
                    'text': time_format(parsed, 'H:i')}
            times.append(timed)
        times.sort(lambda x,y: cmp(x.get('id'), y.get('id')))
        return times

    def cancel(self, id, **kwargs):
        appointment_id = int(id)
        req = self.get_request('appointments/%s' % appointment_id)
        req.get_method = (lambda: 'DELETE')
        try:
            fd = urllib2.urlopen(req)
            none = fd.read()
        except urllib2.HTTPError as e:
            # clicrdv will return a "Bad Request" (HTTP 400) response
            # when it's not possible to remove an appointment
            # (for example because it's too late)
            response = e.read()
            response = json.loads(response)
            return {'success': False, 'error': response}
        return {'success': True}

    def get_data(self, method, **kwargs):
        if not method in ('get_interventionsets', 'get_interventions',
                'get_datetimes', 'get_dates', 'get_times', 'cancel'):
            raise Exception('unknown method')
        call = getattr(self, method)
        return call(**kwargs)


    def create_appointment(self, intervention, websource, data):
        fields = data['fields']
        if intervention:
            intervention = int(intervention)
        else:
            intervention = int(fields.get('clicrdv_intervention_raw'))
        date = fields.get('clicrdv_date_raw')
        time = fields.get('clicrdv_time_raw')
        if not date and not time:
            date = fields.get('clicrdv_datetime_raw')
        else:
            date = date + ' ' + time
        if websource is None:
            websource = self.websource or ''
        appointment = {
            'fiche': {
                'firstname': fields.get('clicrdv_firstname') or '-',
                'lastname': fields.get('clicrdv_lastname') or '-',
                'email': fields.get('clicrdv_email', ''),
                'firstphone': fields.get('clicrdv_firstphone', ''),
                'secondphone': fields.get('clicrdv_secondphone', ''),
             },
             'date': date,
             'intervention_ids': [intervention],
             'websource': websource,
        }
        comments = fields.get('clicrdv_comments') or self.default_comment
        if comments:
            appointment['comments'] = comments
        # optional parameters, if any...
        for fieldname in fields:
            if fieldname.startswith('clicrdv_fiche_'):
                appointment['fiche'][fieldname[14:]] = fields[fieldname] or ''
        req = self.get_request('appointments')
        req.add_data(json.dumps({'appointment': appointment}))
        req.add_header('Content-Type', 'application/json')
        try:
            fd = urllib2.urlopen(req)
        except urllib2.HTTPError, e:
            try:
                error = json.load(e.fp)[0].get('error')
            except:
                error = 'Unknown error (Passerelle)'
            return {
                'success': False,
                'error': error,
            }
        else:
            success = True
            response = json.load(fd)
            appointment_id = response.get('records')[0].get('id')
            return {
                'success': True,
                'appointment_id': appointment_id,
            }

    def post_data(self, data, method, **kwargs):
        if method == 'create':
            intervention = kwargs.get('intervention')
            if intervention:
                intervention = int(intervention)
            websource = kwargs.get('websource')
            return self.create_appointment(intervention, websource, data)
        raise Exception('unknown method')

